]> git.proxmox.com Git - mirror_ubuntu-artful-kernel.git/commitdiff
mm: move MADV_FREE pages into LRU_INACTIVE_FILE list
authorShaohua Li <shli@fb.com>
Wed, 3 May 2017 21:52:29 +0000 (14:52 -0700)
committerLinus Torvalds <torvalds@linux-foundation.org>
Wed, 3 May 2017 22:52:08 +0000 (15:52 -0700)
madv()'s MADV_FREE indicate pages are 'lazyfree'.  They are still
anonymous pages, but they can be freed without pageout.  To distinguish
these from normal anonymous pages, we clear their SwapBacked flag.

MADV_FREE pages could be freed without pageout, so they pretty much like
used once file pages.  For such pages, we'd like to reclaim them once
there is memory pressure.  Also it might be unfair reclaiming MADV_FREE
pages always before used once file pages and we definitively want to
reclaim the pages before other anonymous and file pages.

To speed up MADV_FREE pages reclaim, we put the pages into
LRU_INACTIVE_FILE list.  The rationale is LRU_INACTIVE_FILE list is tiny
nowadays and should be full of used once file pages.  Reclaiming
MADV_FREE pages will not have much interfere of anonymous and active
file pages.  And the inactive file pages and MADV_FREE pages will be
reclaimed according to their age, so we don't reclaim too many MADV_FREE
pages too.  Putting the MADV_FREE pages into LRU_INACTIVE_FILE_LIST also
means we can reclaim the pages without swap support.  This idea is
suggested by Johannes.

This patch doesn't move MADV_FREE pages to LRU_INACTIVE_FILE list yet to
avoid bisect failure, next patch will do it.

The patch is based on Minchan's original patch.

[akpm@linux-foundation.org: coding-style fixes]
Link: http://lkml.kernel.org/r/2f87063c1e9354677b7618c647abde77b07561e5.1487965799.git.shli@fb.com
Signed-off-by: Shaohua Li <shli@fb.com>
Suggested-by: Johannes Weiner <hannes@cmpxchg.org>
Acked-by: Johannes Weiner <hannes@cmpxchg.org>
Acked-by: Minchan Kim <minchan@kernel.org>
Acked-by: Michal Hocko <mhocko@suse.com>
Acked-by: Hillf Danton <hillf.zj@alibaba-inc.com>
Cc: Hugh Dickins <hughd@google.com>
Cc: Rik van Riel <riel@redhat.com>
Cc: Mel Gorman <mgorman@techsingularity.net>
Signed-off-by: Andrew Morton <akpm@linux-foundation.org>
Signed-off-by: Linus Torvalds <torvalds@linux-foundation.org>
include/linux/swap.h
include/linux/vm_event_item.h
mm/huge_memory.c
mm/madvise.c
mm/swap.c
mm/vmstat.c

index 45e91dd6716d89b0ffa8f2d97481d12732d20173..486494e6b2fcdb12692fdd6f57293f5ed3c4ea11 100644 (file)
@@ -279,7 +279,7 @@ extern void lru_add_drain_cpu(int cpu);
 extern void lru_add_drain_all(void);
 extern void rotate_reclaimable_page(struct page *page);
 extern void deactivate_file_page(struct page *page);
-extern void deactivate_page(struct page *page);
+extern void mark_page_lazyfree(struct page *page);
 extern void swap_setup(void);
 
 extern void add_page_to_unevictable_list(struct page *page);
index a80b7b59cf33418811217faca1b9c6b041dad814..d84ae90ccd5c40de69de9217aa7c4e23e4c26be8 100644 (file)
@@ -25,7 +25,7 @@ enum vm_event_item { PGPGIN, PGPGOUT, PSWPIN, PSWPOUT,
                FOR_ALL_ZONES(PGALLOC),
                FOR_ALL_ZONES(ALLOCSTALL),
                FOR_ALL_ZONES(PGSCAN_SKIP),
-               PGFREE, PGACTIVATE, PGDEACTIVATE,
+               PGFREE, PGACTIVATE, PGDEACTIVATE, PGLAZYFREE,
                PGFAULT, PGMAJFAULT,
                PGLAZYFREED,
                PGREFILL,
index 17f6008f282797d9c4b3ef6bf08a55826e02f28c..7309a716b7fc303e79226ea435499ffe3588eec0 100644 (file)
@@ -1564,9 +1564,6 @@ bool madvise_free_huge_pmd(struct mmu_gather *tlb, struct vm_area_struct *vma,
                ClearPageDirty(page);
        unlock_page(page);
 
-       if (PageActive(page))
-               deactivate_page(page);
-
        if (pmd_young(orig_pmd) || pmd_dirty(orig_pmd)) {
                pmdp_invalidate(vma, addr, pmd);
                orig_pmd = pmd_mkold(orig_pmd);
index 7a2abf0127aef7a9d4879278293d8cab766133e1..cf3021b05b320cffe21cebb55ab5a6c53ac736b9 100644 (file)
@@ -411,8 +411,6 @@ static int madvise_free_pte_range(pmd_t *pmd, unsigned long addr,
                        ptent = pte_mkold(ptent);
                        ptent = pte_mkclean(ptent);
                        set_pte_at(mm, addr, pte, ptent);
-                       if (PageActive(page))
-                               deactivate_page(page);
                        tlb_remove_tlb_entry(tlb, pte, addr);
                }
        }
index d8d9ee9e311a6f25520f304a01c77f24a3538584..98d08b4579faa42b8ef43ba4563ed6cb76f19893 100644 (file)
--- a/mm/swap.c
+++ b/mm/swap.c
@@ -46,7 +46,7 @@ int page_cluster;
 static DEFINE_PER_CPU(struct pagevec, lru_add_pvec);
 static DEFINE_PER_CPU(struct pagevec, lru_rotate_pvecs);
 static DEFINE_PER_CPU(struct pagevec, lru_deactivate_file_pvecs);
-static DEFINE_PER_CPU(struct pagevec, lru_deactivate_pvecs);
+static DEFINE_PER_CPU(struct pagevec, lru_lazyfree_pvecs);
 #ifdef CONFIG_SMP
 static DEFINE_PER_CPU(struct pagevec, activate_page_pvecs);
 #endif
@@ -571,20 +571,27 @@ static void lru_deactivate_file_fn(struct page *page, struct lruvec *lruvec,
 }
 
 
-static void lru_deactivate_fn(struct page *page, struct lruvec *lruvec,
+static void lru_lazyfree_fn(struct page *page, struct lruvec *lruvec,
                            void *arg)
 {
-       if (PageLRU(page) && PageActive(page) && !PageUnevictable(page)) {
-               int file = page_is_file_cache(page);
-               int lru = page_lru_base_type(page);
+       if (PageLRU(page) && PageAnon(page) && PageSwapBacked(page) &&
+           !PageUnevictable(page)) {
+               bool active = PageActive(page);
 
-               del_page_from_lru_list(page, lruvec, lru + LRU_ACTIVE);
+               del_page_from_lru_list(page, lruvec,
+                                      LRU_INACTIVE_ANON + active);
                ClearPageActive(page);
                ClearPageReferenced(page);
-               add_page_to_lru_list(page, lruvec, lru);
+               /*
+                * lazyfree pages are clean anonymous pages. They have
+                * SwapBacked flag cleared to distinguish normal anonymous
+                * pages
+                */
+               ClearPageSwapBacked(page);
+               add_page_to_lru_list(page, lruvec, LRU_INACTIVE_FILE);
 
-               __count_vm_event(PGDEACTIVATE);
-               update_page_reclaim_stat(lruvec, file, 0);
+               __count_vm_events(PGLAZYFREE, hpage_nr_pages(page));
+               update_page_reclaim_stat(lruvec, 1, 0);
        }
 }
 
@@ -614,9 +621,9 @@ void lru_add_drain_cpu(int cpu)
        if (pagevec_count(pvec))
                pagevec_lru_move_fn(pvec, lru_deactivate_file_fn, NULL);
 
-       pvec = &per_cpu(lru_deactivate_pvecs, cpu);
+       pvec = &per_cpu(lru_lazyfree_pvecs, cpu);
        if (pagevec_count(pvec))
-               pagevec_lru_move_fn(pvec, lru_deactivate_fn, NULL);
+               pagevec_lru_move_fn(pvec, lru_lazyfree_fn, NULL);
 
        activate_page_drain(cpu);
 }
@@ -648,22 +655,22 @@ void deactivate_file_page(struct page *page)
 }
 
 /**
- * deactivate_page - deactivate a page
+ * mark_page_lazyfree - make an anon page lazyfree
  * @page: page to deactivate
  *
* deactivate_page() moves @page to the inactive list if @page was on the active
* list and was not an unevictable page.  This is done to accelerate the reclaim
* of @page.
  */
-void deactivate_page(struct page *page)
+void mark_page_lazyfree(struct page *page)
 {
-       if (PageLRU(page) && PageActive(page) && !PageUnevictable(page)) {
-               struct pagevec *pvec = &get_cpu_var(lru_deactivate_pvecs);
+       if (PageLRU(page) && PageAnon(page) && PageSwapBacked(page) &&
+           !PageUnevictable(page)) {
+               struct pagevec *pvec = &get_cpu_var(lru_lazyfree_pvecs);
 
                get_page(page);
                if (!pagevec_add(pvec, page) || PageCompound(page))
-                       pagevec_lru_move_fn(pvec, lru_deactivate_fn, NULL);
-               put_cpu_var(lru_deactivate_pvecs);
+                       pagevec_lru_move_fn(pvec, lru_lazyfree_fn, NULL);
+               put_cpu_var(lru_lazyfree_pvecs);
        }
 }
 
@@ -703,7 +710,7 @@ void lru_add_drain_all(void)
                if (pagevec_count(&per_cpu(lru_add_pvec, cpu)) ||
                    pagevec_count(&per_cpu(lru_rotate_pvecs, cpu)) ||
                    pagevec_count(&per_cpu(lru_deactivate_file_pvecs, cpu)) ||
-                   pagevec_count(&per_cpu(lru_deactivate_pvecs, cpu)) ||
+                   pagevec_count(&per_cpu(lru_lazyfree_pvecs, cpu)) ||
                    need_activate_page_drain(cpu)) {
                        INIT_WORK(work, lru_add_drain_per_cpu);
                        queue_work_on(cpu, mm_percpu_wq, work);
index c8d15051616b4b94329af25a914ce833d221302d..828a36ea05848b7258bbde07cae1b9ddb386f00f 100644 (file)
@@ -991,6 +991,7 @@ const char * const vmstat_text[] = {
        "pgfree",
        "pgactivate",
        "pgdeactivate",
+       "pglazyfree",
 
        "pgfault",
        "pgmajfault",
@@ -648,22 +655,22 @@ void deactivate_file_page(struct page *page)
  * @page: page to deactivate
  *
- * deactivate_page() moves @page to the inactive list if @page was on the active
- * list and was not an unevictable page.  This is done to accelerate the reclaim
- * of @page.
+ * mark_page_lazyfree() moves @page to the inactive file list.
+ * This is done to accelerate the reclaim of @page.
  */